import nengo
import numpy as np

from nengo_spa.algebras.base import (
    AbstractAlgebra,
    AbstractSign,
    CommonProperties,
    ElementSidedness,
)
from nengo_spa.networks.circularconvolution import CircularConvolution


class HrrAlgebra(AbstractAlgebra):
    r"""
    Holographic Reduced Representations (HRRs) algebra.

    Uses element-wise addition for superposition, circular convolution for
    binding with an approximate inverse.

    The circular convolution :math:`c` of vectors :math:`a` and :math:`b`
    is given by

    .. math:: c[i] = \sum_j a[j] b[i - j]

    where negative indices on :math:`b` wrap around to the end of the vector.

    This computation can also be done in the Fourier domain,

    .. math:: c = DFT^{-1} ( DFT(a) \odot DFT(b) )

    where :math:`DFT` is the Discrete Fourier Transform operator, and
    :math:`DFT^{-1}` is its inverse.

    Circular convolution as a binding operation is associative, commutative,
    distributive.

    More information on circular convolution as a binding operation can be
    found in [plate2003]_.

    .. [plate2003] Plate, Tony A. Holographic Reduced Representation:
       Distributed Representation for Cognitive Structures. Stanford, CA: CSLI
       Publications, 2003.
    """

    _instance = None

    def __new__(cls):
        if type(cls._instance) is not cls:
            cls._instance = super(HrrAlgebra, cls).__new__(cls)
        return cls._instance

    def is_valid_dimensionality(self, d):
        """
        Checks whether *d* is a valid vector dimensionality.

        For circular convolution all positive numbers are valid
        dimensionalities.

        Parameters
        ----------
        d : int
            Dimensionality

        Returns
        -------
        bool
            *True*, if *d* is a valid vector dimensionality for the use with
            the algebra.
        """
        return d > 0

    def create_vector(self, d, properties, *, rng=None):
        """
        Create a vector fulfilling given properties in the algebra.

        Parameters
        ----------
        d : int
            Vector dimensionality
        properties : set of str
            Definition of properties for the vector to fulfill. Valid set
            elements are constants defined in `.HrrProperties`.
        rng : numpy.random.RandomState, optional
            The random number generator to use to create the vector.

        Returns
        -------
        ndarray
            Random vector with desired properties.
        """
        properties = set(properties)

        if rng is None:
            rng = np.random.RandomState()

        v = rng.randn(d)
        v /= np.linalg.norm(v)

        if HrrProperties.POSITIVE in properties:
            properties.remove(HrrProperties.POSITIVE)
            v = self.abs(v)

        if HrrProperties.UNITARY in properties:
            properties.remove(HrrProperties.UNITARY)
            v = self.make_unitary(v)

        if len(properties) > 0:
            raise ValueError("Invalid properties: " + ", ".join(properties))

        return v

    def make_unitary(self, v):
        fft_val = np.fft.fft(v)
        fft_imag = fft_val.imag
        fft_real = fft_val.real
        fft_norms = np.sqrt(fft_imag**2 + fft_real**2)
        invalid = fft_norms <= 0.0
        fft_val[invalid] = 1.0
        fft_norms[invalid] = 1.0
        fft_unit = fft_val / fft_norms
        return np.array((np.fft.ifft(fft_unit, n=len(v))).real)

    def superpose(self, a, b):
        return a + b

    def bind(self, a, b):
        n = len(a)
        if len(b) != n:
            raise ValueError("Inputs must have same length.")
        return np.fft.irfft(np.fft.rfft(a) * np.fft.rfft(b), n=n)

    def binding_power(self, v, exponent):
        r"""
        Returns the binding power of *v* using the *exponent*.

        The binding power is defined as binding (*exponent*-1) times bindings
        of *v* to itself. Fractional binding powers are supported.

        Note the following special exponents:

        * an exponent of -1 will return the approximate inverse,
        * an exponent of 0 will return the identity vector,
        * and an *exponent* of w1cne will return *v* itself.

        The following relations hold for integer exponents, and for unitary
        vectors:

        * :math:`v^a \circledast v^b = v^{a+b}`,
        * :math:`(v^a)^b = v^{ab}`.

        If :math:`a \geq 0` and :math:`b \geq 0`, then the first relation holds also
        for non-unitary vectors with real exponents.

        Parameters
        ----------
        v : (d,) ndarray
            Vector to bind repeatedly to itself.
        exponent : int or float
            Exponent of the binding power.

        Returns
        -------
        (d,) ndarray
            Binding power of *v*.

        See also
        --------
        .sign
        """
        if int(exponent) != exponent and not self.sign(v).is_positive():
            raise ValueError(
                "Fractional binding powers are only supported for 'positive' vectors."
            )
        if exponent < 0:
            v = self.invert(v)
        return np.fft.irfft(np.fft.rfft(v) ** abs(exponent), n=len(v))

    def invert(self, v, sidedness=ElementSidedness.TWO_SIDED):
        """
        Invert vector *v*.

        This turns circular convolution into circular correlation, meaning that
        ``A*B*~B`` is approximately ``A``.

        Examples
        --------
        For the vector ``[1, 2, 3, 4, 5]``, the inverse is ``[1, 5, 4, 3, 2]``.

        Parameters
        ----------
        v : (d,) ndarray
            Vector to invert.
        sidedness : ElementSidedness, optional
            This argument has no effect because the HRR algebra is commutative
            and the inverse is two-sided.

        Returns
        -------
        (d,) ndarray
            Inverted vector.
        """
        return v[-np.arange(len(v))]

    def get_binding_matrix(self, v, swap_inputs=False):
        D = len(v)
        T = []
        for i in range(D):
            T.append([v[(i - j) % D] for j in range(D)])
        return np.array(T)

    def get_inversion_matrix(self, d, sidedness=ElementSidedness.TWO_SIDED):
        """
        Returns the transformation matrix for inverting a vector.

        Parameters
        ----------
        d : int
            Vector dimensionality (determines the matrix size).
        sidedness : ElementSidedness, optional
            This argument has no effect because the HRR algebra is commutative
            and the inverse is two-sided.

        Returns
        -------
        (d, d) ndarray
            Transformation matrix to invert a vector.
        """
        return np.eye(d)[-np.arange(d)]

    def implement_superposition(self, n_neurons_per_d, d, n):
        node = nengo.Node(size_in=d)
        return node, n * (node,), node

    def implement_binding(self, n_neurons_per_d, d, unbind_left, unbind_right):
        net = CircularConvolution(n_neurons_per_d, d, unbind_left, unbind_right)
        return net, (net.input_a, net.input_b), net.output

    def sign(self, v):
        """
        Returns the HRR sign of *v*.

        See `.AbstractAlgebra.sign` for general information on the notion of a
        sign for algbras, and `.HrrSign` for details specific to HRRs.

        Parameters
        ----------
        v : (d,) ndarray
            Vector to determine sign of.

        Returns
        -------
        HrrSign
            The sign of the input vector.
        """
        dc, nyquist = np.fft.rfft(v)[[0, -1]]
        if len(v) % 2 == 1:
            nyquist = 0
        assert np.isclose(dc.imag, 0) and np.isclose(nyquist.imag, 0)
        return HrrSign(int(np.sign(dc.real)), int(np.sign(nyquist.real)))

    def absorbing_element(self, d, sidedness=ElementSidedness.TWO_SIDED):
        r"""
        Return the standard absorbing element of dimensionality *d*.

        An absorbing element will produce a scaled version of itself when bound
        to another vector. The standard absorbing element is the absorbing
        element with norm 1.

        The absorbing element for circular convolution is the vector
        :math:`(1, 1, \dots, 1)^{\top} / \sqrt{d}`.

        Parameters
        ----------
        d : int
            Vector dimensionality.
        sidedness : ElementSidedness, optional
            This argument has no effect because the HRR algebra is commutative
            and the standard absorbing element is two-sided.

        Returns
        -------
        (d,) ndarray
            Standard absorbing element.
        """
        return np.ones(d) / np.sqrt(d)

    def identity_element(self, d, sidedness=ElementSidedness.TWO_SIDED):
        r"""
        Return the identity element of dimensionality *d*.

        The identity does not change the vector it is bound to.

        The identity element for circular convolution is the vector
        :math:`(1, 0, \dots, 0)^{\top}`.

        Parameters
        ----------
        d : int
            Vector dimensionality.
        sidedness : ElementSidedness, optional
            This argument has no effect because the HRR algebra is commutative
            and the identity is two-sided.

        Returns
        -------
        (d,) ndarray
            Identity element.
        """
        data = np.zeros(d)
        data[0] = 1.0
        return data

    def negative_identity_element(self, d, sidedness=ElementSidedness.TWO_SIDED):
        r"""
        Return the negative identity element of dimensionality *d*.

        The negative identity element for circular convolution is the vector
        :math:`(-1, 0, \dots, 0)^{\top}`.

        Parameters
        ----------
        d : int
            Vector dimensionality.
        sidedness : ElementSidedness, optional
            This argument has no effect because the HRR algebra is commutative
            and the identity is two-sided.

        Returns
        -------
        (d,) ndarray
            Negative identity element.
        """
        return -self.identity_element(d, sidedness)

    def zero_element(self, d, sidedness=ElementSidedness.TWO_SIDED):
        """
        Return the zero element of dimensionality *d*.

        The zero element produces itself when bound to a different vector.
        For circular convolution this is the zero vector.

        Parameters
        ----------
        d : int
            Vector dimensionality.
        sidedness : ElementSidedness, optional
            This argument has no effect because the HRR algebra is commutative
            and the zero element is two-sided.

        Returns
        -------
        (d,) ndarray
            Zero element.
        """
        return np.zeros(d)


class HrrSign(AbstractSign):
    r"""
    Represents a sign in the `.HrrAlgebra`.

    For odd dimensionalities, the sign is equal to the sign of the DC component
    of the Fourier representation of the vector. For even dimensionalities the
    sign is constituted out of the signs of the DC component and Nyquist
    frequency. Thus, for even dimensionalities, there is a total of four
    sub-signs excluding zero. The overall sign is considered positive if the
    DC component is positive and the Nyquist component is non-negative;
    the sign is considered negative if either component is negative; and
    the sign is considered zero if both are zero. Binding two Semantic Pointers
    with the same sub-sign will yield a positive Semantic Pointer. See the table
    below for details.

    .. table:: Resulting Semantic Pointer signs from HRR binding two Semantic
         Pointers. (Only the upper triangle is given as the matrix is
         symmetric.)

        ================== =========== ========== ========== ========== ======
        Sign (DC, Nyquist) \+ (+1, +1) − (+1, -1) − (-1, +1) − (−1, -1) (0, 0)
        ================== =========== ========== ========== ========== ======
        \+ (+1, +1)        \+ (+1, +1) − (+1, -1) − (−1, +1) − (−1, -1) (0, 0)
        − (+1, -1)                     \+ (1, +1) − (−1, -1) − (−1, +1) (0, 0)
        − (−1, +1)                                \+ (1, +1) − (+1, -1) (0, 0)
        − (−1, -1)                                           \+ (1, +1) (0, 0)
        (0, 0)                                                          (0, 0)
        ================== =========== ========== ========== ========== ======

    Parameters
    ----------
    dc_sign : int
        Sign of the DC component.
    nyquist_sign : int
        Sign of the Nyquist frequency component. Will be set to the *dc_sign*
        if zero.
    """

    __slots__ = ["dc_sign", "nyquist_sign"]

    def __init__(self, dc_sign, nyquist_sign):
        if dc_sign == 0 and nyquist_sign != 0:
            raise ValueError(
                "nyquist_sign must be 0 if dc_sign is 0 to constitute a valid "
                "sign in the HrrAlgebra."
            )
        if dc_sign not in (-1, 0, 1):
            raise ValueError("dc_sign must be one of -1, 0, 1")
        if nyquist_sign not in (-1, 0, 1):
            raise ValueError("nyquist_sign must be one of -1, 0, 1")

        self.dc_sign = dc_sign
        self.nyquist_sign = nyquist_sign
        if self.nyquist_sign == 0:
            self.nyquist_sign = self.dc_sign

    def is_positive(self):
        return self.dc_sign > 0 and self.nyquist_sign >= 0

    def is_negative(self):
        return self.dc_sign < 0 or self.nyquist_sign < 0

    def is_indefinite(self):
        return False

    def to_vector(self, d):
        """
        Return the vector in the algebra corresponding to the sign.

        =======  ============  =======================================
        DC sign  Nyquist sign  Vector
        =======  ============  =======================================
         1        1            [ 1,  0, 0, ...] (identity)
         1       -1            [ 0,  1, 0, 0, ...]
        -1        1            [ 0, -1, 0, ...]
        -1       -1            [-1,  0, 0, 0, ...] (negative identity)
         0        0            [ 0,  0, 0, ...] (zero)
        =======  ============  =======================================

        Parameters
        ----------
        d : int
            Vector dimensionality.

        Returns
        -------
        (d,) ndarray
            Vector corresponding to the sign.
        """
        if self.dc_sign == 0:
            return np.zeros(d)

        v = HrrAlgebra().identity_element(d)
        if self.dc_sign * self.nyquist_sign < 0:
            v = np.roll(v, 1)
        return self.dc_sign * v

    def __repr__(self):
        return "{}(dc_sign={}, nyquist_sign={})".format(
            self.__class__.__name__, self.dc_sign, self.nyquist_sign
        )

    def __eq__(self, other):
        if not isinstance(other, HrrSign):
            return False
        return self.dc_sign == other.dc_sign and self.nyquist_sign == other.nyquist_sign


class HrrProperties:
    """Vector properties supported by the `.HrrAlgebra`."""

    UNITARY = CommonProperties.UNITARY
    """A unitary vector does not change the length of a vector it is bound to."""

    POSITIVE = CommonProperties.POSITIVE
    """
    A positive vector does not change the sign of a vector it is bound to.

    A positive vector allows for fractional binding powers.
    """



"""Generators to create vectors with specific properties."""

import numpy as np


def AxisAlignedVectors(d):
    """
    Generator for axis aligned vectors.

    Can yield at most *d* vectors.

    Note that while axis-aligned vectors can be useful for debugging,
    they will not work well with most binding methods for Semantic Pointers.

    Parameters
    ----------
    d : int
        Dimensionality of returned vectors.

    Examples
    --------
    >>> for p in nengo_spa.vector_generation.AxisAlignedVectors(4):
    ...    print(p)
    [1. 0. 0. 0.]
    [0. 1. 0. 0.]
    [0. 0. 1. 0.]
    [0. 0. 0. 1.]
    """
    for v in np.eye(d):
        yield v


class UnitLengthVectors:
    """
    Generator for uniformly distributed unit-length vectors.

    Parameters
    ----------
    d : int
        Dimensionality of returned vectors.
    rng : numpy.random.RandomState, optional
        The random number generator to use to create new vectors.
    """

    def __init__(self, d, rng=None):
        if rng is None:
            rng = np.random.RandomState()

        self.d = d
        self.rng = rng

    def __iter__(self):
        return self

    def __next__(self):
        v = self.rng.randn(self.d)
        v /= np.linalg.norm(v)
        return v

    def next(self):
        return self.__next__()


class UnitaryVectors:
    """
    Generator for unitary vectors (given some binding method).

    Parameters
    ----------
    d : int
        Dimensionality of returned vectors.
    algebra : AbstractAlgebra
        Algebra that defines what vectors are unitary.
    rng : numpy.random.RandomState, optional
        The random number generator to use to create new vectors.
    """

    def __init__(self, d, algebra, rng=None):
        if rng is None:
            rng = np.random.RandomState()

        self.d = d
        self.algebra = algebra
        self.rng = rng

    def __iter__(self):
        return self

    def __next__(self):
        return self.algebra.make_unitary(self.rng.randn(self.d))

    def next(self):
        return self.__next__()


class EquallySpacedPositiveUnitaryHrrVectors:
    """
    Generator for equally spaced positive unitary HRR vectors.

    The vectors produced by this generator lie all on a hyper-circle of
    positive, unitary vectors under the `.HrrAlgebra`. The distance from one
    vector to the next is constant.

    Note that the identity vector is included in the set of returned
    vectors if any of the vectors hits an offset of 0. This might not be desired
    as it will return any vector it is bound to unchanged. Use a non-integer
    *offset* to ensure that the identity vector is not included.

    Parameters
    ----------
    d : int
        Dimensionality of returned vectors.
    n : int
        Number of vectors to fit onto the hyper-circle. At most *n* vectors
        can be returned from the generator.
    offset : float
        Offset of the first returned vector along the hyper-circle. An offset
        of 0 will return the identity vector first. An offset of 1 corresponds
        to the vector when moving a 1/n-th part along the hyper-circle.

    Attributes
    ----------
    vectors : (n, d) ndarray
        All vectors that would be returned by iterating over the generator.
    """

    def __init__(self, *, d, n, offset):
        coefficient_count = (d + 1) // 2
        unity_roots = np.exp(
            2.0j
            * np.pi
            * np.arange(start=coefficient_count, stop=d % 2 - 1, step=-1)
            / coefficient_count
        )
        exponents_offset = coefficient_count / n * offset
        exponents = np.linspace(
            0 + exponents_offset,
            coefficient_count + exponents_offset,
            n,
            endpoint=False,
        )
        self.vectors = np.fft.irfft(unity_roots[None, :] ** exponents[:, None], n=d)

    def __iter__(self):
        return iter(self.vectors)


class OrthonormalVectors:
    """
    Generator for random orthonormal vectors.

    Parameters
    ----------
    d : int
        Dimensionality of returned vectors.
    rng : numpy.random.RandomState, optional
        The random number generator to use to create new vectors.
    """

    def __init__(self, d, rng=None):
        if rng is None:
            rng = np.random.RandomState()

        self.d = d
        self.rng = rng
        self.vectors = []

    def __iter__(self):
        return self

    def __next__(self):
        v = self.rng.randn(self.d)
        i = len(self.vectors)
        if i >= self.d:
            raise StopIteration()
        elif i > 0:
            vectors = np.asarray(self.vectors)
            y = -np.dot(vectors[:, i:], v[i:])
            A = vectors[:i, :i]
            v[:i] = np.linalg.solve(A, y)
        v /= np.linalg.norm(v)
        self.vectors.append(v)
        return v

    def next(self):
        return self.__next__()


class ExpectedUnitLengthVectors:
    r"""
    Generator for vectors with expected unit-length.

    The vectors will be uniformly distributed with an expected norm
    of 1, but each specific pointer may have a length different than 1.
    Specifically each vector component will be normal distributed with mean 0
    and standard deviation :math:`1/\sqrt{d}`.

    Parameters
    ----------
    d : int
        Dimensionality of returned vectors.
    rng : numpy.random.RandomState, optional
        The random number generator to use to create new vectors.
    """

    def __init__(self, d, rng=None):
        if rng is None:
            rng = np.random.RandomState()

        self.d = d
        self.rng = rng

    def __iter__(self):
        return self

    def __next__(self):
        return self.rng.randn(self.d) / np.sqrt(self.d)

    def next(self):
        return self.__next__()


class VectorsWithProperties:
    """
    Generator for vectors with given properties.

    Supported properties depend on the algebra. See the respective algebra's
    :meth:`.AbstractAlgebra.create_vector` method.

    Parameters
    ----------
    d : int
        Dimensionality of returned vectors.
    properties
        Properties that the generated vectors have to fulfill. Details depend
        on the exact algebra.
    algebra : AbstractAlgebra
        Algebra that determines the interpretation of the properties.
    rng : numpy.random.RandomState, optional
        The random number generator to use to create new vectors.
    """

    def __init__(self, d, properties, algebra, *, rng=None):
        if rng is None:
            rng = np.random.RandomState()

        self.d = d
        self.properties = properties
        self.algebra = algebra
        self.rng = rng

    def __iter__(self):
        return self

    def __next__(self):
        return self.algebra.create_vector(self.d, self.properties, rng=self.rng)

    def next(self):
        return self.__next__()



import pytest

from nengo_spa.ast.expr_tree import (
    AttributeAccess,
    BinaryOperator,
    EllipsisLeaf,
    FunctionCall,
    KeywordArgument,
    Leaf,
    UnaryOperator,
    limit_str_length,
)


def test_leaf_str():
    assert str(Leaf("foo")) == "foo"


def test_unary_operator_str():
    assert str(UnaryOperator("~", UnaryOperator("-", Leaf("leaf")))) == "~(-leaf)"
    assert (
        str(UnaryOperator("~", BinaryOperator("+", Leaf("a"), Leaf("b")))) == "~(a + b)"
    )
    assert (
        str(UnaryOperator("~", BinaryOperator("**", Leaf("a"), Leaf("b")))) == "~a ** b"
    )


def test_binary_operator_str():
    a, b, c = Leaf("a"), Leaf("b"), Leaf("c")
    assert str(BinaryOperator("+", a, BinaryOperator("*", b, c))) == "a + b * c"
    assert str(BinaryOperator("*", BinaryOperator("+", a, b), c)) == "(a + b) * c"
    assert str(BinaryOperator("-", BinaryOperator("+", a, b), c)) == "a + b - c"


def test_attribute_access_str():
    assert str(AttributeAccess("attr", Leaf("foo"))) == "foo.attr"
    assert (
        str(AttributeAccess("attr", BinaryOperator("+", Leaf("a"), Leaf("b"))))
        == "(a + b).attr"
    )


def test_function_call_str():
    assert str(FunctionCall((), Leaf("fn"))) == "fn()"
    assert str(FunctionCall(("str1", "str2"), Leaf("fn"))) == "fn(str1, str2)"
    assert (
        str(FunctionCall((UnaryOperator("~", Leaf("foo")),), Leaf("fn"))) == "fn(~foo)"
    )
    assert (
        str(FunctionCall((KeywordArgument("kwarg", "val"),), Leaf("fn")))
        == "fn(kwarg=val)"
    )
    assert (
        str(FunctionCall((), BinaryOperator("+", Leaf("a"), Leaf("b")))) == "(a + b)()"
    )


@pytest.mark.parametrize("op", ["+", "-", "~"])
def test_unary_operator(op):
    leaf = Leaf("leaf")
    assert leaf
    assert str(eval(f"{op}leaf")) == op + "leaf"


@pytest.mark.parametrize("op", ["<<", ">>", "+", "-", "*", "@", "/", "//", "%", "**"])
def test_binary_operator(op):
    lhs, rhs = Leaf("lhs"), Leaf("rhs")
    assert lhs
    assert rhs
    assert str(eval(f"lhs {op} rhs")) == f"lhs {op} rhs"


@pytest.mark.parametrize(
    "expr_tree,max_len,expected",
    [
        (expr_tree, len(str(expr_tree)), expr_tree)
        for expr_tree in (
            Leaf("foo"),
            Leaf("not_too_long"),
            -Leaf("foo"),
            Leaf("a") + Leaf("b"),
            AttributeAccess("attr", Leaf("foo")),
            FunctionCall(("arg",), Leaf("foo")),
        )
    ]
    + [
        (Leaf("foo"), 0, Leaf("foo")),
        (Leaf("too_long"), len("too_long") - 1, EllipsisLeaf()),
        (Leaf("too_long"), 3, EllipsisLeaf()),
        (Leaf("too_long"), 0, EllipsisLeaf()),
        (
            Leaf("varA") + Leaf("varB"),
            len("varA + varB") - 1,
            Leaf("varA") + EllipsisLeaf(),
        ),
        (Leaf("varA") + Leaf("varB"), len("varA + ..."), Leaf("varA") + EllipsisLeaf()),
        (
            Leaf("varA") + Leaf("varB"),
            len("varA + ...") - 1,
            EllipsisLeaf() + EllipsisLeaf(),
        ),
        (Leaf("varA") + Leaf("varB"), 3, EllipsisLeaf()),
        (Leaf("a") + Leaf("b"), len("a + b") - 1, EllipsisLeaf()),
        (
            (Leaf("a") + Leaf("b")) * Leaf("c"),
            len("(a + b) * c") - 1,
            EllipsisLeaf() * Leaf("c"),
        ),
        ((Leaf("a") + Leaf("b")) * Leaf("c"), len("(a + b) * c") - 5, EllipsisLeaf()),
        (
            (Leaf("a") + Leaf("b")) * Leaf("varC"),
            len("(a + b) * varC") - 1,
            (Leaf("a") + Leaf("b")) * EllipsisLeaf(),
        ),
        (
            Leaf("a") * (Leaf("b") + Leaf("c")),
            len("a * (b + c)") - 1,
            Leaf("a") * EllipsisLeaf(),
        ),
        (
            Leaf("a") * (Leaf("b") + Leaf("varC")),
            len("a * (b + varC)") - 1,
            Leaf("a") * (Leaf("b") + EllipsisLeaf()),
        ),
        (
            Leaf("a") + Leaf("a") * Leaf("b") + Leaf("a") * Leaf("b"),
            len("a + a * b + a * b") - 2,
            Leaf("a") + Leaf("a") * Leaf("b") + EllipsisLeaf(),
        ),
        (
            Leaf("a") / (Leaf("a") * Leaf("b")) / (Leaf("a") * Leaf("b")),
            len("a / (a * b) / (a * b)") - 3,
            Leaf("a") / (Leaf("a") * Leaf("b")) / EllipsisLeaf(),
        ),
        (
            AttributeAccess("attribute", Leaf("varA") + Leaf("varB")),
            len("(varA + varB).attribute") - 3,
            AttributeAccess("attribute", EllipsisLeaf("(...)")),
        ),
        (
            AttributeAccess("attribute", Leaf("a")),
            len("a.attribute") - 1,
            EllipsisLeaf(),
        ),
        (FunctionCall([], Leaf("fn_name")), len("...()") - 1, EllipsisLeaf()),
        (
            FunctionCall([], Leaf("fn_name")),
            len("fn_name()") - 1,
            EllipsisLeaf(),
        ),
        (
            FunctionCall([], Leaf("a") + Leaf("b")),
            len("(a + b)()") - 1,
            EllipsisLeaf(),
        ),
        (
            FunctionCall(["arg1", "arg2"], Leaf("fn_name")),
            len("fn_name(arg1, arg2)") - 1,
            FunctionCall(["arg1", EllipsisLeaf()], Leaf("fn_name")),
        ),
        (
            FunctionCall(["arg1", "arg2"], Leaf("fn_name")),
            len("fn_name(arg1, arg2)") - 2,
            FunctionCall([EllipsisLeaf()], Leaf("fn_name")),
        ),
        (
            FunctionCall(["arg1", "arg2"], Leaf("fn_name")),
            len("fn_name(arg1, arg2)") - 8,
            EllipsisLeaf(),
        ),
        (
            FunctionCall(["arg1", "arg2"], Leaf("a") + Leaf("b")),
            len("(a + b)(arg1, arg2)") - 1,
            FunctionCall(["arg1", "..."], Leaf("a") + Leaf("b")),
        ),
        (
            FunctionCall([Leaf("arg1") + Leaf("arg2")], Leaf("fn_name")),
            len("fn_name(arg1 + arg2)") - 1,
            FunctionCall([Leaf("arg1") + EllipsisLeaf()], Leaf("fn_name")),
        ),
        (
            Leaf("var") + FunctionCall([], Leaf("fn_name")),
            len("var + fn_name()") - 1,
            Leaf("var") + Leaf("..."),
        ),
        (
            FunctionCall([KeywordArgument("key", Leaf("value"))], Leaf("fn_name")),
            len("fn_name(key=value)") - 1,
            FunctionCall([KeywordArgument("key", EllipsisLeaf())], Leaf("fn_name")),
        ),
        (
            FunctionCall([KeywordArgument("key", Leaf("value"))], Leaf("fn_name")),
            len("fn_name(key=...)") - 1,
            FunctionCall([EllipsisLeaf()], Leaf("fn_name")),
        ),
    ],
)
def test_limit_str_length(expr_tree, max_len, expected):
    actual = limit_str_length(expr_tree, max_len)
    assert actual == expected, (
        f'When limiting the expression "{str(expr_tree)}" to length {max_len}, '
        f'"{str(expected)}" is expected, but got "{str(actual)}"'
    )



import warnings

import nengo
import numpy as np

from nengo_spa.algebras.base import (
    AbstractAlgebra,
    CommonProperties,
    ElementSidedness,
    GenericSign,
)
from nengo_spa.networks.vtb import VTB


class VtbAlgebra(AbstractAlgebra):
    r"""
    Vector-derived Transformation Binding (VTB) algebra.

    VTB uses elementwise addition for superposition. The binding operation
    :math:`\mathcal{B}(x, y)` is defined as

    .. math::

       \mathcal{B}(x, y) := V_y x = \left[\begin{array}{ccc}
           V_y' &    0 &    0 \\
              0 & V_y' &    0 \\
              0 &    0 & \ddots
           \end{array}\right] x

    with :math:`d'` blocks

    where

    .. math::

       V_y' = d^{\frac{1}{4}} \left[\begin{array}{cccc}
           y_1            & y_2            & \dots  & y_{d'}  \\
           y_{d' + 1}     & y_{d' + 2}     & \dots  & y_{2d'} \\
           \vdots         & \vdots         & \ddots & \vdots  \\
           y_{d - d' + 1} & y_{d - d' + 2} & \dots  & y_d
       \end{array}\right]

    and

    .. math:: d'^2 = d.

    The approximate inverse :math:`y^+` for :math:`y` is permuting the elements
    such that :math:`V_{y^+} = V_y^T`.

    Note that VTB requires the vector dimensionality to be square.

    The VTB binding operation is neither associative nor commutative.
    Furthermore, there are right inverses and identities only.
    By transposing the :math:`V_y` matrix, the closely related `.TvtbAlgebra`
    (Transposed VTB) algebra is obtained which does have two-sided identities
    and inverses.

    Additional information about VTB can be found in

    * `Gosmann, Jan, and Chris Eliasmith (2019). Vector-derived transformation binding:
      an improved binding operation for deep symbol-like processing in
      neural networks. Neural computation 31.5, 849-869.
      <https://direct.mit.edu/neco/article/31/5/849/8469/Vector-Derived-Transformation-Binding-An-Improved>`_
    * `Jan Gosmann (2018). An Integrated Model of Context, Short-Term, and
      Long-Term Memory. UWSpace. <https://uwspace.uwaterloo.ca/handle/10012/13498>`_

    .. seealso::
        `.TvtbAlgebra`
    """

    _instance = None

    def __new__(cls):
        if type(cls._instance) is not cls:
            cls._instance = super(VtbAlgebra, cls).__new__(cls)
        return cls._instance

    def is_valid_dimensionality(self, d):
        """
        Checks whether *d* is a valid vector dimensionality.

        For VTB all square numbers are valid dimensionalities.

        Parameters
        ----------
        d : int
            Dimensionality

        Returns
        -------
        bool
            *True*, if *d* is a valid vector dimensionality for the use with
            the algebra.
        """
        if d < 1:
            return False
        sub_d = np.sqrt(d)
        return sub_d * sub_d == d

    def _get_sub_d(self, d):
        sub_d = int(np.sqrt(d))
        if sub_d * sub_d != d:
            raise ValueError("Vector dimensionality must be a square number.")
        return sub_d

    def create_vector(self, d, properties, *, rng=None):
        """
        Create a vector fulfilling given properties in the algebra.

        Creating positive vectors requires SciPy.

        Parameters
        ----------
        d : int
            Vector dimensionality
        properties : set of str
            Definition of properties for the vector to fulfill. Valid set
            elements are constants defined in `.VtbProperties`.
        rng : numpy.random.RandomState, optional
            The random number generator to use to create the vector.

        Returns
        -------
        ndarray
            Random vector with desired properties.
        """
        properties = set(properties)

        if rng is None:
            rng = np.random.RandomState()

        if {VtbProperties.UNITARY, VtbProperties.POSITIVE} <= properties:
            properties -= {VtbProperties.UNITARY, VtbProperties.POSITIVE}
            v = self.identity_element(d, sidedness=ElementSidedness.RIGHT)
            warnings.warn(
                "The only positive unitary vector in VTB is the identity. "
                "Use the identity directly to avoid this warning."
            )
        elif VtbProperties.UNITARY in properties:
            properties.remove(VtbProperties.UNITARY)
            v = self.make_unitary(rng.randn(d))
        elif VtbProperties.POSITIVE in properties:
            try:
                from scipy.linalg import sqrtm
            except ImportError as err:
                # From Python 3.6 onward we get a ModuleNotFoundError. We want
                # to re-raise the same type to be as specific as possible.
                raise type(err)(
                    "Creating positive VTB vectors requires SciPy to be available.",
                    name=err.name,
                    path=err.path,
                )

            properties.remove(VtbProperties.POSITIVE)
            sub_d = self._get_sub_d(d)
            v = rng.randn(d)
            v /= np.linalg.norm(v)
            mat = v.reshape((sub_d, sub_d))
            v = sqrtm(np.dot(mat, mat.T)).flatten()
        else:
            v = rng.randn(d)
            v /= np.linalg.norm(v)

        if len(properties) > 0:
            raise ValueError("Invalid properties: " + ", ".join(properties))

        return v

    def make_unitary(self, v):
        sub_d = self._get_sub_d(len(v))
        m = np.array(v.reshape((sub_d, sub_d)))
        for i in range(1, sub_d):
            y = -np.dot(m[:i, i:], m[i, i:])
            A = m[:i, :i]
            m[i, :i] = np.linalg.solve(A, y)
        m /= np.linalg.norm(m, axis=1)[:, None]
        m /= np.sqrt(sub_d)
        return m.flatten()

    def superpose(self, a, b):
        return a + b

    def bind(self, a, b):
        d = len(a)
        if len(b) != d:
            raise ValueError("Inputs must have same length.")
        m = self.get_binding_matrix(b)
        return np.dot(m, a)

    def invert(self, v, sidedness=ElementSidedness.TWO_SIDED):
        """
        Invert vector *v*.

        A vector bound to its inverse will result in the identity vector.

        VTB has a right inverse only.

        .. deprecated:: 1.2.0
           Calling this method with the default
           ``sidedness=ElementSidedness.TWO_SIDED`` returns the right inverse
           for backwards compatibility, but has been deprecated and will be
           removed in the next major release.

        Parameters
        ----------
        v : (d,) ndarray
            Vector to invert.
        sidedness : ElementSidedness
            Must be set to `.ElementSidedness.RIGHT`.

        Returns
        -------
        (d,) ndarray
            Right inverse of vector.
        """
        if sidedness is ElementSidedness.LEFT:
            raise NotImplementedError("VtbAlgebra does not have a left inverse.")
        if sidedness is ElementSidedness.TWO_SIDED:
            warnings.warn(
                DeprecationWarning(
                    "VtbAlgebra does not have a two-sided inverse, returning "
                    "the right inverse instead. Please change your code to "
                    "request the right inverse explicitly with "
                    "`sidedness=ElementSidedness.RIGHT`."
                )
            )

        sub_d = self._get_sub_d(len(v))
        return v.reshape((sub_d, sub_d)).T.flatten()

    def binding_power(self, v, exponent):
        r"""
        Returns the binding power of *v* using the *exponent*.

        The binding power is defined as binding (*exponent*-1) times bindings
        of *v* to itself.

        Fractional binding powers are supported for "positive" vectors if SciPy
        is available.

        Note the following special exponents:

        * an exponent of -1 will return the inverse,
        * an exponent of 0 will return the identity vector,
        * and an *exponent* of 1 will return *v* itself.

        Be aware that the binding power for the VTB algebra does *not* satisfy
        the usual properties of exponentiation:

        * :math:`\mathcal{B}(v^a, v^b) = v^{a+b}` does *not* hold,
        * :math:`(v^a)^b = v^{ab}` does *not* hold.

        Parameters
        ----------
        v : (d,) ndarray
            Vector to bind repeatedly to itself.
        exponent : int or float
            Exponent of the binding power.

        Returns
        -------
        (d,) ndarray
            Binding power of *v*.

        See also
        --------
        .sign
        """

        try:
            from scipy.linalg import fractional_matrix_power
        except ImportError as err:
            if int(exponent) == exponent:
                exponent = int(exponent)
                # Provide fallback for integer-only powers

                def fractional_matrix_power(m, exp):
                    power = np.eye(len(m))
                    for _ in range(exp):
                        power = np.dot(power, m)
                    return power

            else:
                # From Python 3.6 onward we get a ModuleNotFoundError. We want
                # to re-raise the same type to be as specific as possible.
                raise type(err)(
                    "Fractional VTB binding powers require SciPy to be available.",
                    name=err.name,
                    path=err.path,
                )

        if exponent == 0:
            return self.identity_element(len(v), sidedness=ElementSidedness.RIGHT)

        if int(exponent) != exponent and not self.sign(v).is_positive():
            raise ValueError(
                "Fractional binding powers are only supported for 'positive' vectors."
            )

        if exponent < 0:
            v = self.invert(v, sidedness=ElementSidedness.RIGHT)

        sub_d = self._get_sub_d(len(v))
        power = fractional_matrix_power(
            v.reshape((sub_d, sub_d)) * np.sqrt(sub_d), abs(exponent) - 1
        ).flatten() / np.sqrt(sub_d)
        assert np.allclose(power.imag, 0)

        return self.bind(v, power.real)

    def get_binding_matrix(self, v, swap_inputs=False):
        sub_d = self._get_sub_d(len(v))
        m = np.sqrt(sub_d) * np.kron(np.eye(sub_d), v.reshape((sub_d, sub_d)))
        if swap_inputs:
            m = np.dot(self.get_swapping_matrix(len(v)), m)
        return m

    def get_swapping_matrix(self, d):
        """
        Get matrix to swap operands in bound state.

        Parameters
        ----------
        d : int
            Dimensionality of vector.

        Returns
        -------
        (d, d) ndarry
            Matrix to multiply with a vector to switch left and right operand
            in bound state.
        """
        return self.get_inversion_matrix(d, sidedness=ElementSidedness.RIGHT)

    def get_inversion_matrix(self, d, sidedness=ElementSidedness.TWO_SIDED):
        """
        Returns the transformation matrix for inverting a vector.

        VTB has a right inverse only.

        .. deprecated:: 1.2.0
           Calling this method with the default
           ``sidedness=ElementSidedness.TWO_SIDED`` returns the right
           transformation matrix for the right inverse for backwards
           compatibility, but has been deprecated and will be removed in the
           next major release.

        Parameters
        ----------
        d : int
            Vector dimensionality.
        sidedness : ElementSidedness
            Must be set to `.ElementSidedness.RIGHT`.

        Returns
        -------
        (d, d) ndarray
            Transformation matrix to invert a vector.
        """
        if sidedness is ElementSidedness.LEFT:
            raise NotImplementedError("VtbAlgebra does not have a left inverse.")
        if sidedness is ElementSidedness.TWO_SIDED:
            warnings.warn(
                DeprecationWarning(
                    "VtbAlgebra does not have a two-sided inverse, returning "
                    "the right inverse instead. Please change your code to "
                    "request the right inverse explicitly with "
                    "`sidedness=ElementSidedness.RIGHT`."
                )
            )

        sub_d = self._get_sub_d(d)
        return np.eye(d).reshape(d, sub_d, sub_d).T.reshape(d, d)

    def implement_superposition(self, n_neurons_per_d, d, n):
        node = nengo.Node(size_in=d)
        return node, n * (node,), node

    def implement_binding(self, n_neurons_per_d, d, unbind_left, unbind_right):
        net = VTB(n_neurons_per_d, d, unbind_left, unbind_right)
        return net, (net.input_left, net.input_right), net.output

    def sign(self, v):
        """See `.AbstractAlgebra.sign`."""
        m = self.get_binding_matrix(v)
        if not np.allclose(m, m.T):
            return VtbSign(None)
        eigenvalues = np.linalg.eigvalsh(m)
        if np.all(eigenvalues > 0):
            return VtbSign(1)
        elif np.all(eigenvalues < 0):
            return VtbSign(-1)
        elif np.allclose(eigenvalues, 0):
            return VtbSign(0)
        else:
            return VtbSign(None)

    def abs(self, v):
        """See `.AbstractAlgebra.abs`."""
        # No inverse of sign required because in VTB sign vectors are their
        # own inverse.
        return self.bind(v, self.sign(v).to_vector(len(v)))

    def absorbing_element(self, d, sidedness=ElementSidedness.TWO_SIDED):
        """
        VTB has no absorbing element except the zero vector.

        Always raises a `NotImplementedError`.
        """
        raise NotImplementedError("VtbAlgebra does not have any absorbing elements.")

    def identity_element(self, d, sidedness=ElementSidedness.TWO_SIDED):
        """
        Return the identity element of dimensionality *d*.

        VTB has a right identity only.

        .. deprecated:: 1.2.0
           Calling this method with the default
           ``sidedness=ElementSidedness.TWO_SIDED`` returns the right identity
           for backwards compatibility, but has been deprecated and will be
           removed in the next major release.

        Parameters
        ----------
        d : int
            Vector dimensionality.
        sidedness : ElementSidedness
            Must be set to `.ElementSidedness.RIGHT`.

        Returns
        -------
        (d,) ndarray
            Right identity element.
        """
        if sidedness is ElementSidedness.LEFT:
            raise NotImplementedError("VtbAlgebra does not have a left identity.")
        if sidedness is ElementSidedness.TWO_SIDED:
            warnings.warn(
                DeprecationWarning(
                    "VtbAlgebra does not have a two-sided identity, returning "
                    "the right identity instead. Please change your code to "
                    "request the right identity explicitly with "
                    "`sidedness=ElementSidedness.RIGHT`."
                )
            )

        sub_d = self._get_sub_d(d)
        return (np.eye(sub_d) / d**0.25).flatten()

    def negative_identity_element(self, d, sidedness=ElementSidedness.TWO_SIDED):
        r"""
        Return the negative identity element of dimensionality *d*.

        VTB has a right negative identity only.

        Parameters
        ----------
        d : int
            Vector dimensionality.
        sidedness : ElementSidedness, optional
            Must be set to `.ElementSidedness.RIGHT`.

        Returns
        -------
        (d,) ndarray
            Negative identity element.
        """
        if sidedness is not ElementSidedness.RIGHT:
            raise NotImplementedError("VTB only has a right negative identity.")
        return -self.identity_element(d, sidedness)

    def zero_element(self, d, sidedness=ElementSidedness.TWO_SIDED):
        """
        Return the zero element of dimensionality *d*.

        The zero element produces itself when bound to a different vector.
        For VTB this is the zero vector.

        Parameters
        ----------
        d : int
            Vector dimensionality.
        sidedness : ElementSidedness, optional
            This argument has no effect because the zero element of the VTB
            algebra is two-sided.

        Returns
        -------
        (d,) ndarray
            Zero element.
        """
        return np.zeros(d)


class VtbSign(GenericSign):
    """
    Represents a sign in the `.VtbAlgebra`.

    The sign depends on the symmetry and positive/negative definiteness of the
    binding matrix derived from the vector. For all non-symmetric matrices, the
    sign is indefinite. It is also indefinite, if the matrices' eigenvalues have
    different signs. A symmetric, positive (negative) definite binding matrix
    corresponds to a positive (negative) sign (equivalent to all eigenvalues
    being greater than 0, respectively lower than 0). If all eigenvalues are
    equal to 0, the sign is also 0.
    """

    def to_vector(self, d):
        if self.sign is None:
            raise NotImplementedError(
                "There is no vector corresponding to an indefinite sign."
            )
        elif self.sign > 0:
            return VtbAlgebra().identity_element(d, sidedness=ElementSidedness.RIGHT)
        elif self.sign < 0:
            return VtbAlgebra().negative_identity_element(
                d, sidedness=ElementSidedness.RIGHT
            )
        elif self.sign == 0:
            return VtbAlgebra().zero_element(d)
        else:
            raise AssertionError(
                "Invalid value for sign, this code should be unreachable."
            )


class VtbProperties:
    """Vector properties supported by the `.VtbAlgebra`."""

    UNITARY = CommonProperties.UNITARY
    """A unitary vector does not change the length of a vector it is bound to."""

    POSITIVE = CommonProperties.POSITIVE
    """
    A positive vector does not change the sign of a vector it is bound to.

    A positive vector allows for fractional binding powers.
    """



"""Algebras define the specific superposition and (un)binding operations."""

from .base import AbstractAlgebra, CommonProperties, ElementSidedness
from .hrr_algebra import HrrAlgebra, HrrProperties
from .tvtb_algebra import TvtbAlgebra, TvtbProperties
from .vtb_algebra import VtbAlgebra, VtbProperties



import numpy as np
import pytest

from nengo_spa.algebras import CommonProperties, HrrAlgebra
from nengo_spa.semantic_pointer import SemanticPointer
from nengo_spa.vector_generation import (
    AxisAlignedVectors,
    EquallySpacedPositiveUnitaryHrrVectors,
    ExpectedUnitLengthVectors,
    OrthonormalVectors,
    UnitaryVectors,
    UnitLengthVectors,
    VectorsWithProperties,
)


def test_axis_aligned_pointers():
    for i, p in enumerate(AxisAlignedVectors(5)):
        assert np.all(p == np.eye(5)[i])


def test_unit_length_pointers(rng):
    g = UnitLengthVectors(64, rng=rng)
    for i in range(10):
        assert np.allclose(np.linalg.norm(next(g)), 1.0)


def test_unitary_pointers(rng):
    algebra = HrrAlgebra()
    g = UnitaryVectors(64, algebra, rng=rng)
    a = SemanticPointer(next(g), algebra=algebra)
    b = SemanticPointer(next(g), algebra=algebra)
    c = SemanticPointer(next(g), algebra=algebra)
    assert np.allclose(a.compare(c), (a * b).compare(c * b))


@pytest.mark.parametrize(
    "constructor_kwargs,expected",
    [
        ({"d": 16, "n": 8, "offset": 0}, np.eye(16)[::2]),
        ({"d": 16, "n": 8, "offset": 2}, np.roll(np.eye(16)[::2], -2, axis=0)),
        (
            {"d": 16, "n": 8, "offset": 0.3},
            HrrAlgebra().binding_power(np.eye(16)[2], 0.3),
        ),
        ({"d": 16, "n": 20, "offset": 1}, None),
        ({"d": 9, "n": 9, "offset": 0}, np.eye(9)),
        ({"d": 9, "n": 9, "offset": 2}, np.roll(np.eye(9), -2, axis=0)),
        (
            {"d": 9, "n": 9, "offset": 0.3},
            HrrAlgebra().binding_power(np.eye(9)[1], 0.3),
        ),
        ({"d": 9, "n": 20, "offset": 1}, None),
    ],
)
def test_equally_spaced_positive_unitary_hrr_vectors(constructor_kwargs, expected):
    g = EquallySpacedPositiveUnitaryHrrVectors(**constructor_kwargs)

    assert np.all(
        np.array(list(g)) == g.vectors
    ), "Iterating should yield `vectors` property."
    assert len(g.vectors) == constructor_kwargs["n"], "Should produce `n` vectors."

    similarities = np.array(
        [
            np.dot(g.vectors[i], g.vectors[(i + 1) % len(g.vectors)])
            for i in range(len(g.vectors))
        ]
    )
    assert np.allclose(
        similarities, similarities[0]
    ), "Similarities of neighbouring pairs should all be the same."
    assert np.all(similarities < 1.0), "Vectors should not be the same."


def test_orthonormal_pointers(rng):
    g = OrthonormalVectors(32, rng=rng)
    vectors = np.array(list(g))
    assert len(vectors) == 32
    assert np.allclose(np.dot(vectors.T, vectors), np.eye(32))


def test_expected_unit_length_vectors(rng):
    g = ExpectedUnitLengthVectors(64, rng=rng)
    assert np.abs(np.mean([np.linalg.norm(next(g)) for i in range(50)]) - 1.0) < 0.1


def test_vectors_with_properties(rng):
    algebra = HrrAlgebra()
    g = VectorsWithProperties(
        64, {CommonProperties.POSITIVE, CommonProperties.UNITARY}, algebra, rng=rng
    )
    v = next(g)
    sqrt_v = algebra.binding_power(v, 0.5)
    assert np.allclose(v, algebra.bind(sqrt_v, sqrt_v)), "is positive"
    assert np.allclose(algebra.make_unitary(v), v)


@pytest.mark.parametrize(
    "vg",
    (
        AxisAlignedVectors,
        ExpectedUnitLengthVectors,
        OrthonormalVectors,
        UnitLengthVectors,
        lambda d: UnitaryVectors(d, algebra=HrrAlgebra()),
        lambda d: VectorsWithProperties(
            d, {CommonProperties.POSITIVE}, algebra=HrrAlgebra()
        ),
    ),
)
def test_instantiation_without_rng(vg):
    d = 64
    assert len(next(vg(d))) == d


@pytest.mark.parametrize(
    "vg",
    (
        AxisAlignedVectors,
        ExpectedUnitLengthVectors,
        OrthonormalVectors,
        UnitLengthVectors,
        lambda d: UnitaryVectors(d, algebra=HrrAlgebra()),
        lambda d: VectorsWithProperties(
            d, {CommonProperties.POSITIVE}, algebra=HrrAlgebra()
        ),
    ),
)
def test_iter(vg):
    d = 64
    x = vg(d)
    assert iter(x) is x



"""Representation of simple expression trees."""

from collections import namedtuple

# Reference precedence table:
# https://docs.python.org/3/reference/expressions.html
precedence = {
    op: p
    for p, operators in enumerate(
        (
            {":="},
            {"lambda"},
            {"if", "else"},
            {"or"},
            {"and"},
            {"not x"},
            {"in", "not in", "is", "is not", "<", "<=", ">", ">=", "!=", "=="},
            {"|"},
            {"^"},
            {"&"},
            {"<<", ">>"},
            {"+", "-"},
            {"*", "@", "/", "//", "%"},
            {"+x", "-x", "~x"},
            {"**"},
            {"await x"},
            {"x[index]", "x[index:index]", "x(arguments...)", "x.attribute"},
            {
                "(expressions...)",
                "[expressions...]",
                "{key: value...}",
                "{expressions...}",
            },
        )
    )
    for op in operators
}


class Node(namedtuple("Node", ["value", "precedence", "children"])):
    def __invert__(self):
        return UnaryOperator("~", self)

    def __neg__(self):
        return UnaryOperator("-", self)

    def __pos__(self):
        return UnaryOperator("+", self)

    def __add__(self, other):
        return BinaryOperator("+", self, other)

    def __radd__(self, other):
        return BinaryOperator("+", other, self)

    def __sub__(self, other):
        return BinaryOperator("-", self, other)

    def __rsub__(self, other):
        return BinaryOperator("-", other, self)

    def __mul__(self, other):
        return BinaryOperator("*", self, other)

    def __rmul__(self, other):
        return BinaryOperator("*", other, self)

    def __truediv__(self, other):
        return BinaryOperator("/", self, other)

    def __floordiv__(self, other):
        return BinaryOperator("//", self, other)

    def __mod__(self, other):
        return BinaryOperator("%", self, other)

    def __matmul__(self, other):
        return BinaryOperator("@", self, other)

    def __rmatmul__(self, other):
        return BinaryOperator("@", other, self)

    def __pow__(self, other):
        return BinaryOperator("**", self, other)

    def __lshift__(self, other):
        return BinaryOperator("<<", self, other)

    def __rshift__(self, other):
        return BinaryOperator(">>", self, other)


class Leaf(Node):
    __slots__ = ()

    def __new__(cls, value):
        return super().__new__(cls, value, precedence["(expressions...)"], tuple())

    def __str__(self):
        return str(self.value)


class EllipsisLeaf(Leaf):
    __slots__ = ()

    def __new__(cls, value="..."):
        return super().__new__(cls, value)


class UnaryOperator(Node):
    __slots__ = ()

    def __new__(cls, value, child):
        return super().__new__(cls, value, precedence[value + "x"], (child,))

    def __str__(self):
        child = self.children[0]
        operand = str(child)
        if self.precedence >= child.precedence:
            operand = f"({operand})"
        return self.value + operand


class BinaryOperator(Node):
    __slots__ = ()

    def __new__(cls, value, lhs, rhs):
        return super().__new__(cls, value, precedence[value], (lhs, rhs))

    def __str__(self):
        lhs, rhs = self.children
        if self.precedence > lhs.precedence:
            lhs = f"({lhs})"
        if self.precedence >= rhs.precedence:
            rhs = f"({rhs})"
        return f"{lhs} {self.value} {rhs}"

    @property
    def lhs(self):
        return self.children[0]

    @property
    def rhs(self):
        return self.children[1]


class AttributeAccess(Node):
    __slots__ = ()

    def __new__(cls, value, child):
        return super().__new__(cls, value, precedence["x.attribute"], (child,))

    def __str__(self):
        child = self.children[0]
        operand = str(child)
        if self.precedence > child.precedence:
            operand = f"({operand})"
        return operand + "." + self.value


class FunctionCall(Node):
    __slots__ = ()

    def __new__(cls, arguments, child):
        return super().__new__(
            cls,
            [arg if isinstance(arg, Node) else Leaf(arg) for arg in arguments],
            precedence["x(arguments...)"],
            (child,),
        )

    def __str__(self):
        child = self.children[0]
        operand = str(child)
        if self.precedence > child.precedence:
            operand = f"({operand})"
        return f"{operand}({', '.join(str(arg) for arg in self.value)})"


class KeywordArgument(Node):
    __slots__ = ()

    def __new__(cls, value, child):
        return super().__new__(cls, value, 0, (child,))

    def __str__(self):
        return f"{self.value}={str(self.children[0])}"


class _LimitStrLengthVisitor:
    # General idea of the algorithm: Do an in-order traversal and keep track
    # how much characters we may still use (max_len). When the limit is
    # exceeded, start replacing nodes with ellipsis and re-adjust max_len to
    # account for the characters saved through the replacement.

    def __init__(self, max_len):
        self.max_len = max_len

    def visit_Leaf(self, node):
        if len(node.value) <= 3 or len(node.value) <= self.max_len:
            self.max_len -= len(node.value)
            return node
        else:
            self.max_len -= 3
            return EllipsisLeaf()

    def visit_EllipsisLeaf(self, node):
        self.max_len -= len(str(node))
        return node

    def visit_UnaryOperator(self, node):
        self.max_len -= len(node.value)
        if node.precedence >= node.children[0].precedence:
            self.max_len -= 2
        return UnaryOperator(node.value, self.visit_node(node.children[0]))

    def visit_BinaryOperator(self, node):
        initial_max_len = self.max_len

        # Operator + 2 spaces + reserve 3 characters for the second operand, so
        # that it can be at least turned into an ellipsis
        self.max_len -= len(node.value) + 2 + 3
        lhs = self.visit_node(node.lhs)
        self.max_len += 3  # return reserved characters for the second operand
        # Adjust for parenthesis due to operator precedence
        if node.precedence > lhs.precedence:
            self.max_len -= 2
        if node.precedence >= node.rhs.precedence:
            self.max_len -= 2

        rhs = self.visit_node(node.rhs)
        if self.max_len < 0:
            if node.precedence >= node.rhs.precedence:
                self.max_len += 2
            self.max_len += len(str(rhs)) - 3
            rhs = EllipsisLeaf()
        if self.max_len >= 0:
            return BinaryOperator(node.value, lhs, rhs)
        else:
            self.max_len = initial_max_len - 3
            return EllipsisLeaf()

    def visit_AttributeAccess(self, node):
        initial_max_len = self.max_len
        self.max_len -= len(node.value) + 1
        child = self.visit_node(node.children[0])
        if node.precedence > child.precedence:
            self.max_len -= 2
        if self.max_len < 0:
            self.max_len += len(str(child)) - 5
            child = EllipsisLeaf("(...)")
        if self.max_len >= 0:
            return AttributeAccess(node.value, child)
        else:
            self.max_len = initial_max_len - 3
            return EllipsisLeaf()

    def visit_FunctionCall(self, node):
        initial_max_len = self.max_len
        child = self.visit_node(node.children[0])
        self.max_len -= 2  # parenthesis invoking call
        if node.precedence > child.precedence:
            self.max_len -= 2  # parenthesis around function

        args = []
        for i, arg in enumerate(node.value):
            if i > 0:
                self.max_len -= 2  # comma + space argument separator
            arg = self.visit_node(arg)
            if i > 0 and self.max_len < 0:
                self.max_len += (
                    len(str(args[-1])) - 3
                )  # replacement of previous arg with ellipsis
                self.max_len += (
                    len(str(arg)) + 2
                )  # adjust for dropped, but processed arg and separator
                args[-1] = EllipsisLeaf()
            else:
                args.append(arg)

        if self.max_len >= 0:
            return FunctionCall(args, child)
        else:
            self.max_len = initial_max_len - 3
            return EllipsisLeaf()

    def visit_KeywordArgument(self, node):
        initial_max_len = self.max_len
        self.max_len -= len(str(node.value)) + 1  # 1 for assignment (=)
        child = self.visit_node(node.children[0])
        if self.max_len >= 0:
            return KeywordArgument(node.value, child)
        else:
            self.max_len = initial_max_len - 3
            return EllipsisLeaf()

    def visit_node(self, node):
        return getattr(self, "visit_" + type(node).__name__)(node)


def limit_str_length(expr_tree, max_len):
    """Returns a modified expression tree with a string length limited to approximately
    ``max_len``."""
    return _LimitStrLengthVisitor(max_len).visit_node(expr_tree)



import numpy as np

from nengo_spa.typechecks import is_integer, is_number


def test_is_integer():
    assert is_integer(1)
    assert is_integer(np.array(1))
    assert not is_integer(np.array([1]))
    assert not is_integer(1.0)
    assert not is_integer(np.array(1.0))
    assert not is_integer(np.array([1.0]))
    assert not is_integer(object())


def test_is_number():
    assert is_number(1)
    assert is_number(np.array(1))
    assert is_number(1.0)
    assert is_number(np.array(1.0))
    assert not is_number(np.array([1.0]))
    assert not is_number(np.array([1]))
    assert not is_number(object())



import numpy as np
import pytest

from nengo_spa.algebras.vtb_algebra import VtbAlgebra, VtbProperties, VtbSign
from nengo_spa.semantic_pointer import SemanticPointer
from nengo_spa.vector_generation import UnitLengthVectors


def test_is_singleton():
    assert VtbAlgebra() is VtbAlgebra()


def test_is_valid_dimensionality():
    assert not VtbAlgebra().is_valid_dimensionality(-1)
    assert not VtbAlgebra().is_valid_dimensionality(0)
    assert not VtbAlgebra().is_valid_dimensionality(15)
    assert not VtbAlgebra().is_valid_dimensionality(24)
    assert VtbAlgebra().is_valid_dimensionality(1)
    assert VtbAlgebra().is_valid_dimensionality(16)
    assert VtbAlgebra().is_valid_dimensionality(25)


def test_get_swapping_matrix(rng):
    gen = UnitLengthVectors(64, rng=rng)
    a = SemanticPointer(next(gen), algebra=VtbAlgebra()).v
    b = SemanticPointer(next(gen), algebra=VtbAlgebra()).v

    m = VtbAlgebra().get_swapping_matrix(64)
    assert np.allclose(VtbAlgebra().bind(a, b), np.dot(m, VtbAlgebra().bind(b, a)))


@pytest.mark.parametrize(
    "sign",
    [VtbSign(-1), VtbSign(1)],
)
def test_sign_and_abs(sign):
    pytest.importorskip("scipy")
    d = 16
    algebra = VtbAlgebra()
    abs_v = algebra.create_vector(d, {VtbProperties.POSITIVE})
    v = algebra.bind(sign.to_vector(d), abs_v)
    assert algebra.sign(v) == sign
    assert np.allclose(algebra.abs(v), abs_v)


def test_sign_repr():
    sign = VtbSign(-1)
    assert eval(repr(sign)) == sign


def test_create_positive_vector(rng):
    pytest.importorskip("scipy")
    algebra = VtbAlgebra()
    v = algebra.create_vector(16, {VtbProperties.POSITIVE}, rng=rng)
    assert len(v) == 16
    assert algebra.sign(v).is_positive()


def test_create_unitary_vector(rng):
    algebra = VtbAlgebra()
    v = algebra.create_vector(16, {VtbProperties.UNITARY}, rng=rng)
    assert len(v) == 16
    assert np.allclose(algebra.make_unitary(v), v)


@pytest.mark.filterwarnings("ignore:.*only positive unitary vector")
def test_create_positive_unitary_vector(rng):
    algebra = VtbAlgebra()
    v = algebra.create_vector(16, {VtbProperties.UNITARY, VtbProperties.POSITIVE})
    assert len(v) == 16
    assert algebra.sign(v).is_positive()
    assert np.allclose(algebra.make_unitary(v), v)


def test_create_vector_with_invalid_property():
    with pytest.raises(ValueError):
        VtbAlgebra().create_vector(16, "foo")


def test_fractional_binding_power_of_non_positive_vector_raises(rng):
    pytest.importorskip("scipy")
    algebra = VtbAlgebra()
    v = algebra.bind(
        algebra.create_vector(16, {VtbProperties.POSITIVE}, rng=rng),
        VtbSign(-1).to_vector(16),
    )
    assert algebra.sign(v).is_negative()
    with pytest.raises(ValueError):
        algebra.binding_power(v, 0.5)


def test_fractional_binding_power_with_positive_exponent_lower_1(rng):
    pytest.importorskip("scipy")

    algebra = VtbAlgebra()
    v = algebra.create_vector(16, {VtbProperties.POSITIVE}, rng=rng)
    assert np.allclose(v, algebra.binding_power(algebra.binding_power(v, 2.5), 1 / 2.5))



import warnings

import numpy as np
import pytest

from nengo_spa.algebras.base import AbstractAlgebra, CommonProperties, ElementSidedness
from nengo_spa.algebras.vtb_algebra import VtbAlgebra
from nengo_spa.vector_generation import UnitLengthVectors


@pytest.mark.parametrize("d", [16, 25])
def test_make_unitary(algebra, d, rng):
    a = next(UnitLengthVectors(d, rng=rng))
    b = algebra.make_unitary(a)
    for i in range(3):
        assert np.allclose(1, np.linalg.norm(a))
        assert np.allclose(1, np.linalg.norm(b))
        a = algebra.bind(a, b)
        assert np.allclose(1, np.linalg.norm(a))


def test_superpose(algebra, rng):
    gen = UnitLengthVectors(16, rng=rng)
    a = next(gen)
    b = next(gen)
    # Orthogonalize
    b -= np.dot(a, b) * a
    b /= np.linalg.norm(b)

    r = algebra.superpose(a, b)
    for v in (a, b):
        assert np.dot(v, r / np.linalg.norm(r)) > 0.6


@pytest.mark.parametrize("d", [25, 36])
@pytest.mark.parametrize("sidedness", ElementSidedness)
def test_binding_and_invert(algebra, d, sidedness, rng):
    dissimilarity_passed = 0
    unbinding_passed = 0
    try:
        for i in range(10):
            gen = UnitLengthVectors(d, rng=rng)
            a = next(gen)
            b = next(gen)

            binding_side = sidedness
            if sidedness is ElementSidedness.TWO_SIDED:
                binding_side = (
                    ElementSidedness.LEFT if i % 1 == 0 else ElementSidedness.RIGHT
                )

            with warnings.catch_warnings():
                warnings.simplefilter("error", DeprecationWarning)
                if binding_side is ElementSidedness.LEFT:
                    bound = algebra.bind(b, a)
                    r = algebra.bind(algebra.invert(b, sidedness=sidedness), bound)
                elif binding_side is ElementSidedness.RIGHT:
                    bound = algebra.bind(a, b)
                    r = algebra.bind(bound, algebra.invert(b, sidedness=sidedness))
                else:
                    raise AssertionError("Invalid binding_side value.")

            for v in (a, b):
                dissimilarity_passed += np.dot(v, bound / np.linalg.norm(b)) < 0.7
            unbinding_passed += np.dot(a, r / np.linalg.norm(r)) > 0.6

        assert dissimilarity_passed >= 2 * 8
        assert unbinding_passed >= 7
    except (NotImplementedError, DeprecationWarning):
        pass


@pytest.mark.parametrize("d", [25, 36])
@pytest.mark.filterwarnings("ignore:.*sidedness:DeprecationWarning")
def test_integer_binding_power(algebra, d, rng):
    v = algebra.create_vector(d, set(), rng=rng)
    assert np.allclose(
        algebra.binding_power(v, -2), algebra.bind(algebra.invert(v), algebra.invert(v))
    )
    assert np.allclose(algebra.binding_power(v, -1), algebra.invert(v))
    assert np.allclose(algebra.binding_power(v, 0), algebra.identity_element(d))
    assert np.allclose(algebra.binding_power(v, 1), v)
    assert np.allclose(algebra.binding_power(v, 2), algebra.bind(v, v))
    assert np.allclose(algebra.binding_power(v, 3), algebra.bind(algebra.bind(v, v), v))


@pytest.mark.parametrize("d", [25, 36])
def test_integer_binding_is_consistent_with_base_implementation(algebra, d, rng):
    v = algebra.create_vector(d, set(), rng=rng)

    try:
        for exponent in range(-2, 4):
            assert np.allclose(
                algebra.binding_power(v, exponent),
                AbstractAlgebra.binding_power(algebra, v, exponent),
            )

        with pytest.raises(ValueError, match="only supports integer binding powers"):
            AbstractAlgebra.binding_power(algebra, v, 0.5)
    except NotImplementedError:
        pytest.skip()


@pytest.mark.parametrize("d", [16, 25])
@pytest.mark.filterwarnings("ignore:.*sidedness:DeprecationWarning")
def test_fractional_binding_power(algebra, d, rng):
    pytest.importorskip("scipy")
    v = algebra.create_vector(d, {CommonProperties.POSITIVE}, rng=rng)

    try:
        sqrt_v = algebra.binding_power(v, 0.5)
        assert np.allclose(algebra.bind(sqrt_v, sqrt_v), v)

        inv_sqrt_v = algebra.binding_power(v, -0.5)
        assert np.allclose(algebra.bind(inv_sqrt_v, inv_sqrt_v), algebra.invert(v))
    except ValueError as err:
        assert "only supports integer binding powers" in str(err)


def test_dimensionality_mismatch_exception(algebra):
    with pytest.raises(ValueError):
        algebra.bind(np.ones(16), np.ones(25))
    with pytest.raises(ValueError):
        algebra.superpose(np.ones(16), np.ones(25))


def test_get_binding_matrix(algebra, rng):
    gen = UnitLengthVectors(16, rng=rng)
    a = next(gen)
    b = next(gen)

    m = algebra.get_binding_matrix(b)
    assert np.allclose(algebra.bind(a, b), np.dot(m, a))

    m = algebra.get_binding_matrix(b, swap_inputs=True)
    assert np.allclose(algebra.bind(b, a), np.dot(m, a))


@pytest.mark.filterwarnings("ignore:.*sidedness:DeprecationWarning")
@pytest.mark.parametrize("sidedness", ElementSidedness)
def test_get_inversion_matrix(algebra, sidedness, rng):
    a = next(UnitLengthVectors(16, rng=rng))
    try:
        m = algebra.get_inversion_matrix(16, sidedness=sidedness)
        assert np.allclose(algebra.invert(a, sidedness=sidedness), np.dot(m, a))
    except NotImplementedError:
        pass


@pytest.mark.parametrize("sidedness", ElementSidedness)
def test_absorbing_element(algebra, sidedness, rng):
    a = next(UnitLengthVectors(16, rng=rng))
    try:
        with warnings.catch_warnings(record=True) as caught_warnings:
            warnings.simplefilter("always", DeprecationWarning)
            p = algebra.absorbing_element(16)
    except NotImplementedError:
        pass
    else:
        is_deprecated = len(caught_warnings) > 0 and any(
            issubclass(w.category, DeprecationWarning) for w in caught_warnings
        )
        if (
            sidedness in (ElementSidedness.LEFT, ElementSidedness.TWO_SIDED)
            and not is_deprecated
        ):
            r = algebra.bind(p, a)
            r /= np.linalg.norm(r)
            assert np.allclose(p, r) or np.allclose(p, -r)
        if sidedness in (ElementSidedness.RIGHT, ElementSidedness.TWO_SIDED):
            r = algebra.bind(a, p)
            r /= np.linalg.norm(r)
            assert np.allclose(p, r) or np.allclose(p, -r)


@pytest.mark.parametrize("sidedness", ElementSidedness)
def test_identity_element(algebra, sidedness, rng):
    a = next(UnitLengthVectors(16, rng=rng))
    try:
        with warnings.catch_warnings(record=True) as caught_warnings:
            warnings.simplefilter("always", DeprecationWarning)
            p = algebra.identity_element(16)
    except NotImplementedError:
        pass
    else:
        is_deprecated = len(caught_warnings) > 0 and any(
            issubclass(w.category, DeprecationWarning) for w in caught_warnings
        )
        if (
            sidedness in (ElementSidedness.LEFT, ElementSidedness.TWO_SIDED)
            and not is_deprecated
        ):
            assert np.allclose(algebra.bind(p, a), a)
        if sidedness in (ElementSidedness.RIGHT, ElementSidedness.TWO_SIDED):
            assert np.allclose(algebra.bind(a, p), a)


@pytest.mark.parametrize("sidedness", ElementSidedness)
def test_negative_identity_element(algebra, sidedness, rng):
    try:
        x = next(UnitLengthVectors(16, rng=rng))
        a = algebra.abs(x)
        with warnings.catch_warnings(record=True) as caught_warnings:
            warnings.simplefilter("always", DeprecationWarning)
            p = algebra.negative_identity_element(16)
    except NotImplementedError:
        pass
    else:
        is_deprecated = len(caught_warnings) > 0 and any(
            issubclass(w.category, DeprecationWarning) for w in caught_warnings
        )
        if (
            sidedness in (ElementSidedness.LEFT, ElementSidedness.TWO_SIDED)
            and not is_deprecated
        ):
            b = algebra.bind(p, a)
            assert np.allclose(algebra.abs(b), a)
            assert algebra.sign(b).is_negative()
        if sidedness in (ElementSidedness.RIGHT, ElementSidedness.TWO_SIDED):
            b = algebra.bind(a, p)
            assert np.allclose(algebra.abs(b), a)
            assert algebra.sign(b).is_negative()


@pytest.mark.parametrize("sidedness", ElementSidedness)
def test_zero_element(algebra, sidedness, rng):
    a = next(UnitLengthVectors(16, rng=rng))
    try:
        with warnings.catch_warnings(record=True) as caught_warnings:
            warnings.simplefilter("always", DeprecationWarning)
            p = algebra.zero_element(16)
    except NotImplementedError:
        pass
    else:
        assert np.all(p == 0.0)
        is_deprecated = len(caught_warnings) > 0 and any(
            issubclass(w.category, DeprecationWarning) for w in caught_warnings
        )
        if (
            sidedness in (ElementSidedness.LEFT, ElementSidedness.TWO_SIDED)
            and not is_deprecated
        ):
            assert np.allclose(algebra.bind(a, p), 0.0)
        if sidedness in (ElementSidedness.RIGHT, ElementSidedness.TWO_SIDED):
            assert np.allclose(algebra.bind(p, a), 0.0)


def test_isinstance_check(algebra):
    assert isinstance(algebra, AbstractAlgebra)


class DummyAlgebra:
    def is_valid_dimensionality(self, d):
        pass

    def create_vector(self, d, properties, *, rng=None):
        pass

    def make_unitary(self, v):
        pass

    def superpose(self, a, b):
        pass

    def bind(self, a, b):
        pass

    def binding_power(self, v, exponent):
        pass

    def invert(self, v):
        pass

    def get_binding_matrix(self, v, swap_inputs=False):
        pass

    def get_inversion_matrix(self, d):
        pass

    def implement_superposition(self, n_neurons_per_d, d, n):
        pass

    def implement_binding(self, n_neurons_per_d, d, unbind_left, unbind_right):
        pass

    def sign(self, v):
        pass

    def abs(self, v):
        pass

    def absorbing_element(self, d):
        pass

    def identity_element(self, d):
        pass

    def negative_identity_element(self, d):
        pass

    def zero_element(self, d):
        pass


@pytest.mark.filterwarnings("ignore:.*do not rely on pure duck-typing")
def test_isinstance_ducktyping_check():
    assert isinstance(DummyAlgebra(), AbstractAlgebra)


@pytest.mark.parametrize(
    "element,check_property",
    [
        ("identity", "is_positive"),
        ("zero", "is_zero"),
        ("negative_identity", "is_negative"),
    ],
)
@pytest.mark.parametrize("sidedness", ElementSidedness)
@pytest.mark.filterwarnings("ignore:.*sidedness:DeprecationWarning")
def test_sign(algebra, element, check_property, sidedness):
    try:
        v = getattr(algebra, element + "_element")(16, sidedness)
        assert getattr(algebra.sign(v), check_property)()
    except NotImplementedError:
        pass


@pytest.mark.parametrize("d", [16, 25])
@pytest.mark.parametrize("sidedness", ElementSidedness)
def test_abs(algebra, d, sidedness):
    try:
        neg_v = algebra.negative_identity_element(d, sidedness)
        assert algebra.sign(neg_v).is_negative()
        v = algebra.abs(neg_v)
        assert algebra.sign(v).is_positive()
        assert np.allclose(v, algebra.identity_element(d, sidedness))
        assert np.allclose(algebra.abs(v), v)  # idempotency
    except NotImplementedError:
        pass



import numpy as np
import pytest

from nengo_spa.algebras.tvtb_algebra import TvtbAlgebra, TvtbProperties, TvtbSign


def test_is_singleton():
    assert TvtbAlgebra() is TvtbAlgebra()


def test_is_valid_dimensionality():
    assert not TvtbAlgebra().is_valid_dimensionality(-1)
    assert not TvtbAlgebra().is_valid_dimensionality(0)
    assert not TvtbAlgebra().is_valid_dimensionality(15)
    assert not TvtbAlgebra().is_valid_dimensionality(24)
    assert TvtbAlgebra().is_valid_dimensionality(1)
    assert TvtbAlgebra().is_valid_dimensionality(16)
    assert TvtbAlgebra().is_valid_dimensionality(25)


@pytest.mark.parametrize(
    "sign",
    [TvtbSign(-1), TvtbSign(1)],
)
def test_sign_and_abs(sign):
    pytest.importorskip("scipy")
    d = 16
    algebra = TvtbAlgebra()
    abs_v = algebra.create_vector(d, {TvtbProperties.POSITIVE})
    v = algebra.bind(sign.to_vector(d), abs_v)
    assert algebra.sign(v) == sign
    assert np.allclose(algebra.abs(v), abs_v)


def test_sign_repr():
    sign = TvtbSign(-1)
    assert eval(repr(sign)) == sign


def test_create_positive_vector(rng):
    pytest.importorskip("scipy")
    algebra = TvtbAlgebra()
    v = algebra.create_vector(16, {TvtbProperties.POSITIVE}, rng=rng)
    assert len(v) == 16
    assert algebra.sign(v).is_positive()


def test_create_unitary_vector(rng):
    algebra = TvtbAlgebra()
    v = algebra.create_vector(16, {TvtbProperties.UNITARY}, rng=rng)
    assert len(v) == 16
    assert np.allclose(algebra.make_unitary(v), v)


@pytest.mark.filterwarnings("ignore:.*only positive unitary vector")
def test_create_positive_unitary_vector(rng):
    algebra = TvtbAlgebra()
    v = algebra.create_vector(16, {TvtbProperties.UNITARY, TvtbProperties.POSITIVE})
    assert len(v) == 16
    assert algebra.sign(v).is_positive()
    assert np.allclose(algebra.make_unitary(v), v)


def test_create_vector_with_invalid_property():
    with pytest.raises(ValueError):
        TvtbAlgebra().create_vector(16, "foo")


def test_additional_integer_binding_power_properties(rng):
    algebra = TvtbAlgebra()
    v = algebra.create_vector(16, {TvtbProperties.UNITARY}, rng=rng)

    assert np.allclose(
        algebra.bind(algebra.binding_power(v, 2), algebra.binding_power(v, 3)),
        algebra.binding_power(v, 5),
    )
    assert np.allclose(
        algebra.binding_power(algebra.binding_power(v, 2), 3),
        algebra.binding_power(v, 6),
    )

    assert np.allclose(
        algebra.bind(algebra.binding_power(v, 2), algebra.binding_power(v, -4)),
        algebra.binding_power(v, -2),
    )
    assert np.allclose(
        algebra.binding_power(algebra.binding_power(v, -2), 3),
        algebra.binding_power(v, -6),
    )


@pytest.mark.filterwarnings("ignore:.*only positive unitary vector")
def test_additional_fractional_binding_power_properties(rng):
    pytest.importorskip("scipy")
    algebra = TvtbAlgebra()
    v = algebra.create_vector(
        16, {TvtbProperties.POSITIVE, TvtbProperties.UNITARY}, rng=rng
    )

    assert np.allclose(
        algebra.bind(algebra.binding_power(v, 2.2), algebra.binding_power(v, 3.3)),
        algebra.binding_power(v, 5.5),
    )
    assert np.allclose(
        algebra.bind(algebra.binding_power(v, 2.2), algebra.binding_power(v, -4.4)),
        algebra.binding_power(v, -2.2),
    )


def test_fractional_binding_power_of_non_positive_vector_raises(rng):
    pytest.importorskip("scipy")
    algebra = TvtbAlgebra()
    v = algebra.bind(
        algebra.create_vector(16, {TvtbProperties.POSITIVE}, rng=rng),
        TvtbSign(-1).to_vector(16),
    )
    assert algebra.sign(v).is_negative()
    with pytest.raises(ValueError):
        algebra.binding_power(v, 0.5)



import warnings
from abc import ABC, ABCMeta
from enum import Enum


class ElementSidedness(Enum):
    """The side in a binary operation for which a special element's properties hold."""

    LEFT = "left"
    RIGHT = "right"
    TWO_SIDED = "two-sided"


class _DuckTypedABCMeta(ABCMeta):
    def __instancecheck__(cls, instance):
        if super().__instancecheck__(instance):
            return True

        for member in dir(cls):
            if member.startswith("_"):
                continue
            if not hasattr(instance, member) or not hasattr(
                getattr(instance, member), "__self__"
            ):
                return False
        warnings.warn(
            DeprecationWarning(
                f"Please do not rely on pure duck-typing for {cls.__name__}. "
                f"Explicitly register your class {instance.__class__.__name__} "
                f"as a virtual subclass of {cls.__name__} or derive from it."
            )
        )
        return True


class AbstractAlgebra(metaclass=_DuckTypedABCMeta):
    """
    Abstract base class for algebras.

    Custom algebras can be defined by implementing the interface of this abstract
    base class.
    """

    def is_valid_dimensionality(self, d):
        """
        Checks whether *d* is a valid vector dimensionality.

        Parameters
        ----------
        d : int
            Dimensionality

        Returns
        -------
        bool
            *True*, if *d* is a valid vector dimensionality for the use with
            the algebra.
        """
        raise NotImplementedError()

    def create_vector(self, d, properties, *, rng=None):
        """
        Create a vector fulfilling given properties in the algebra.

        Valid properties and combinations thereof depend on the concrete
        algebra. It is suggested that the *properties* is either a *set* of
        *str* (if order does not matter) or a *list* of *str* (if order does
        matter). Use the constants defined in `.CommonProperties` where
        appropriate.

        Parameters
        ----------
        d : int
            Vector dimensionality
        properties
            Definition of properties for the vector to fulfill. Type and
            specification format depend on the concrete algbra, but it is
            suggested to use either a *set* or *list* of *str* (depending on
            whether order of properties matters) utilizing the constants defined
            in `.CommonProperties` where applicable.
        rng : numpy.random.RandomState, optional
            The random number generator to use to create the vector.

        Returns
        -------
        ndarray
            Random vector with desired properties.
        """
        raise NotImplementedError()

    def make_unitary(self, v):
        """
        Returns a unitary vector based on the vector *v*.

        A unitary vector does not change the length of a vector it is bound to.

        Parameters
        ----------
        v : (d,) ndarray
            Vector to base unitary vector on.

        Returns
        -------
        ndarray
            Unitary vector.
        """
        raise NotImplementedError()

    def superpose(self, a, b):
        """
        Returns the superposition of *a* and *b*.

        This is commonly elementwise addition.

        Parameters
        ----------
        a : (d,) ndarray
            Left operand in superposition.
        b : (d,) ndarray
            Right operand in superposition.

        Returns
        -------
        (d,) ndarray
            Superposed vector.
        """
        raise NotImplementedError()

    def bind(self, a, b):
        """
        Returns the binding of *a* and *b*.

        The resulting vector should in most cases be dissimilar to both inputs.

        Parameters
        ----------
        a : (d,) ndarray
            Left operand in binding.
        b : (d,) ndarray
            Right operand in binding.

        Returns
        -------
        (d,) ndarray
            Bound vector.
        """
        raise NotImplementedError()

    def binding_power(self, v, exponent):
        """
        Returns the binding power of *v* using the *exponent*.

        For a positive *exponent*, the binding power is defined as binding
        (*exponent*-1) times bindings of *v* to itself. For a negative
        *exponent*, the binding power is the approximate inverse bound to itself
        according to the prior definition. Depending on the algebra, fractional
        exponents might be valid or return a *ValueError*, if not. Usually, a
        fractional binding power will require that *v* has a positive sign.

        Note the following special exponents:

        * an exponent of -1 will return the approximate inverse,
        * an exponent of 0 will return the identity vector,
        * and an *exponent* of 1 will return *v* itself.

        The default implementation supports integer exponents only and will
        apply the `.bind` method multiple times. It requires the algebra to have
        a left identity.

        Parameters
        ----------
        v : (d,) ndarray
            Vector to bind repeatedly to itself.
        exponent : int or float
            Exponent of the binding power.

        Returns
        -------
        (d,) ndarray
            Binding power of *v*.

        See also
        --------
        AbstractAlgebra.sign
        """

        if not int(exponent) == exponent:
            raise ValueError(
                "{} only supports integer binding powers.".format(
                    self.__class__.__name__
                )
            )
        exponent = int(exponent)

        power = self.identity_element(len(v), sidedness=ElementSidedness.LEFT)
        for _ in range(abs(exponent)):
            power = self.bind(power, v)

        if exponent < 0:
            power = self.invert(power)

        return power

    def invert(self, v, sidedness=ElementSidedness.TWO_SIDED):
        """
        Invert vector *v*.

        A vector bound to its inverse will result in the identity vector.

        Some algebras might not have an inverse only on specific sides. In that
        case a *NotImplementedError* may be raised for non-existing inverses.

        Parameters
        ----------
        v : (d,) ndarray
            Vector to invert.
        sidedness : ElementSidedness, optional
            Side in the binding operation on which the returned value acts as
            inverse.

        Returns
        -------
        (d,) ndarray
            Inverted vector.
        """
        raise NotImplementedError()

    def get_binding_matrix(self, v, swap_inputs=False):
        """
        Returns the transformation matrix for binding with a fixed vector.

        Parameters
        ----------
        v : (d,) ndarray
            Fixed vector to derive binding matrix for.
        swap_inputs : bool, optional
            By default the matrix will be such that *v* becomes the *right*
            operand in the binding. By setting *swap_inputs*, the matrix will
            be such that *v* becomes the *left* operand. For binding operations
            that are commutative (such as circular convolution), this has no
            effect.

        Returns
        -------
        (d, d) ndarray
            Transformation matrix to perform binding with *v*.
        """
        raise NotImplementedError()

    def get_inversion_matrix(self, d, sidedness=ElementSidedness.TWO_SIDED):
        """
        Returns the transformation matrix for inverting a vector.

        Some algebras might not have an inverse only on specific sides. In that
        case a *NotImplementedError* may be raised for non-existing inverses.

        Parameters
        ----------
        d : int
            Vector dimensionality (determines the matrix size).
        sidedness : ElementSidedness, optional
            Side in the binding operation on which a transformed vectors acts as
            inverse.

        Returns
        -------
        (d, d) ndarray
            Transformation matrix to invert a vector.
        """
        raise NotImplementedError()

    def implement_superposition(self, n_neurons_per_d, d, n):
        """
        Implement neural network for superposing vectors.

        Parameters
        ----------
        n_neurons_per_d : int
            Neurons to use per dimension.
        d : int
            Dimensionality of the vectors.
        n : int
            Number of vectors to superpose in the network.

        Returns
        -------
        tuple
            Tuple *(net, inputs, output)* where *net* is the implemented
            `nengo.Network`, *inputs* a sequence of length *n* of inputs to the
            network, and *output* the network output.
        """
        raise NotImplementedError()

    def implement_binding(self, n_neurons_per_d, d, unbind_left, unbind_right):
        """
        Implement neural network for binding vectors.

        Parameters
        ----------
        n_neurons_per_d : int
            Neurons to use per dimension.
        d : int
            Dimensionality of the vectors.
        unbind_left : bool
            Whether the left input should be unbound from the right input.
        unbind_right : bool
            Whether the right input should be unbound from the left input.

        Returns
        -------
        tuple
            Tuple *(net, inputs, output)* where *net* is the implemented
            `nengo.Network`, *inputs* a sequence of the left and the right
            input in that order, and *output* the network output.
        """
        raise NotImplementedError()

    def sign(self, v):
        """
        Returns the sign of *v* defined by the algebra.

        The exact definition of the sign depends on the concrete algebra, but
        should be analogous to the sign of a (complex) number in so far that
        binding two vectors with the same sign produces a "positive" vector.
        There might, however, be multiple types of negative signs, where binding
        vectors with different types of negative signs will produce another
        "negative" vector.

        Furthermore, if the algebra supports fractional binding powers, it
        should do so for all "non-negative" vectors, but not "negative" vectors.

        If an algebra does not have the notion of a sign, it may raise a
        :py:class:`NotImplementedError`.

        Parameters
        ----------
        v : (d,) ndarray
            Vector to determine sign of.

        Returns
        -------
        AbstractSign
            The sign of the input vector.

        See Also
        --------
        AbstractAlgebra.abs
        """
        raise NotImplementedError()

    def abs(self, v):
        """
        Returns the absolute vector of *v* defined by the algebra.

        The exact definition of "absolute vector" may depend on the concrete
        algebra. It should be a "positive" vector (see `.sign`) that relates
        to the input vector.

        The default implementation requires that the possible signs of the
        algebra correspond to actual vectors within the algebra. It will bind
        the inverse of the sign vector (from the left side) to the vector *v*.

        If an algebra does not have the notion of a sign or absolute vector,
        it may raise a :py:class:`NotImplementedError`.

        Parameters
        ----------
        v : (d,) ndarray
            Vector to obtain the absolute vector of.

        Returns
        -------
        (d,) ndarray
            The absolute vector relating to the input vector.
        """
        return self.bind(self.invert(self.sign(v).to_vector(len(v))), v)

    def absorbing_element(self, d, sidedness=ElementSidedness.TWO_SIDED):
        """
        Return the standard absorbing element of dimensionality *d*.

        An absorbing element will produce a scaled version of itself when bound
        to another vector. The standard absorbing element is the absorbing
        element with norm 1.

        Some algebras might not have an absorbing element other than the zero
        vector. In that case a *NotImplementedError* may be raised.

        Parameters
        ----------
        d : int
            Vector dimensionality.
        sidedness : ElementSidedness, optional
            Side in the binding operation on which the element absorbs.

        Returns
        -------
        (d,) ndarray
            Standard absorbing element.
        """
        raise NotImplementedError()

    def identity_element(self, d, sidedness=ElementSidedness.TWO_SIDED):
        """
        Return the identity element of dimensionality *d*.

        The identity does not change the vector it is bound to.

        Some algebras might not have an identity element. In that case a
        *NotImplementedError* may be raised.

        Parameters
        ----------
        d : int
            Vector dimensionality.
        sidedness : ElementSidedness, optional
            Side in the binding operation on which the element acts as identity.

        Returns
        -------
        (d,) ndarray
            Identity element.
        """
        raise NotImplementedError()

    def negative_identity_element(self, d, sidedness=ElementSidedness.TWO_SIDED):
        """
        Returns the negative identity element of dimensionality *d*.

        The negative identity only changes the sign of the vector it is bound to.

        Some algebras might not have a negative identity element (or even the
        notion of a sign). In that case a :py:class`NotImplementedError` may be
        raised.

        Parameters
        ----------
        d : int
            Vector dimensionality.
        sidedness : ElementSidedness, optional
            Side in the binding operation on which the element acts as negative
            identity.

        Returns
        -------
        (d,) ndarray
            Negative identity element.
        """
        raise NotImplementedError()

    def zero_element(self, d, sidedness=ElementSidedness.TWO_SIDED):
        """
        Return the zero element of dimensionality *d*.

        The zero element produces itself when bound to a different vector.
        Usually this will be the zero vector.

        Some algebras might not have a zero element. In that case a
        *NotImplementedError* may be raised.

        Parameters
        ----------
        d : int
            Vector dimensionality.
        sidedness : ElementSidedness, optional
            Side in the binding operation on which the element acts as zero.

        Returns
        -------
        (d,) ndarray
            Zero element.
        """
        raise NotImplementedError()


class AbstractSign(ABC):
    """Abstract base class for implementing signs for an algebra."""

    def is_positive(self):
        """Return whether the sign is positive."""
        raise NotImplementedError()

    def is_negative(self):
        """Return whether the sign is negative."""
        raise NotImplementedError()

    def is_zero(self):
        """Return whether the sign neither positive nor negative (i.e. zero), but
        definite."""
        return not (self.is_positive() or self.is_negative() or self.is_indefinite())

    def is_indefinite(self):
        """Return whether the sign is neither positive nor negative nor zero."""
        raise NotImplementedError()

    def to_vector(self, d):
        """
        Return the vector in the algebra corresponding to the sign.

        Parameters
        ----------
        d : int
            Vector dimensionality.

        Returns
        -------
        (d,) ndarray
            Vector corresponding to the sign.
        """
        raise NotImplementedError()


class GenericSign(AbstractSign):
    """
    A generic sign implementation.

    Parameters
    ----------
    sign : -1, 0, 1, None
        The represented sign. *None* is used for an indefinite sign.
    """

    def __init__(self, sign):
        if sign not in (-1, 0, 1, None):
            raise ValueError("sign must be one of -1, 0, 1, None")
        self.sign = sign

    def is_positive(self):
        return not self.is_indefinite() and self.sign > 0

    def is_negative(self):
        return not self.is_indefinite() and self.sign < 0

    def is_zero(self):
        return not self.is_indefinite() and self.sign == 0

    def is_indefinite(self):
        return self.sign is None

    def __repr__(self):
        return "{}(sign={})".format(self.__class__.__name__, self.sign)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.sign == other.sign


class CommonProperties:
    """
    Definition of constants for common properties of vectors in an algebra.

    Use these for best interoperability between algebras.
    """

    UNITARY = "unitary"
    """A unitary vector does not change the length of a vector it is bound to."""

    POSITIVE = "positive"
    """
    A positive vector does not change the sign of a vector it is bound to.

    A positive vector allows for fractional binding powers.
    """



import numpy as np
import pytest

from nengo_spa.algebras.hrr_algebra import HrrAlgebra, HrrProperties, HrrSign
from nengo_spa.vector_generation import UnitLengthVectors


def test_is_singleton():
    assert HrrAlgebra() is HrrAlgebra()


@pytest.mark.parametrize(
    "sign,d",
    [
        (HrrSign(1, 0), 15),
        (HrrSign(-1, 0), 15),
        (HrrSign(1, 1), 16),
        (HrrSign(1, -1), 16),
        (HrrSign(-1, 1), 16),
        (HrrSign(-1, -1), 16),
    ],
)
def test_sign_and_abs(sign, d):
    algebra = HrrAlgebra()
    abs_v = algebra.abs(next(UnitLengthVectors(d)))
    v = algebra.bind(sign.to_vector(d), abs_v)
    assert algebra.sign(v) == sign
    assert np.allclose(algebra.abs(v), abs_v)


def test_create_positive_vector(rng):
    algebra = HrrAlgebra()
    v = algebra.create_vector(16, {HrrProperties.POSITIVE}, rng=rng)
    assert len(v) == 16
    assert algebra.sign(v).is_positive()


def test_create_unitary_vector(rng):
    algebra = HrrAlgebra()
    v = algebra.create_vector(16, {HrrProperties.UNITARY}, rng=rng)
    assert len(v) == 16
    assert np.allclose(algebra.make_unitary(v), v)


def test_create_positive_unitary_vector(rng):
    algebra = HrrAlgebra()
    v = algebra.create_vector(16, {HrrProperties.UNITARY, HrrProperties.POSITIVE})
    assert len(v) == 16
    assert algebra.sign(v).is_positive()
    assert np.allclose(algebra.make_unitary(v), v)


def test_create_vector_with_invalid_property():
    with pytest.raises(ValueError):
        HrrAlgebra().create_vector(16, "foo")


def test_additional_integer_binding_power_properties(rng):
    algebra = HrrAlgebra()
    v = algebra.create_vector(16, {HrrProperties.UNITARY}, rng=rng)

    assert np.allclose(
        algebra.bind(algebra.binding_power(v, 2), algebra.binding_power(v, 3)),
        algebra.binding_power(v, 5),
    )
    assert np.allclose(
        algebra.binding_power(algebra.binding_power(v, 2), 3),
        algebra.binding_power(v, 6),
    )

    assert np.allclose(
        algebra.bind(algebra.binding_power(v, 2), algebra.binding_power(v, -4)),
        algebra.binding_power(v, -2),
    )
    assert np.allclose(
        algebra.binding_power(algebra.binding_power(v, -2), 3),
        algebra.binding_power(v, -6),
    )


@pytest.mark.filterwarnings("ignore:.*only positive unitary vector")
def test_additional_fractional_binding_power_properties(rng):
    algebra = HrrAlgebra()
    v = algebra.create_vector(
        16, {HrrProperties.POSITIVE, HrrProperties.UNITARY}, rng=rng
    )

    assert np.allclose(
        algebra.bind(algebra.binding_power(v, 2.2), algebra.binding_power(v, 3.3)),
        algebra.binding_power(v, 5.5),
    )
    assert np.allclose(
        algebra.bind(algebra.binding_power(v, 2.2), algebra.binding_power(v, -4.4)),
        algebra.binding_power(v, -2.2),
    )

    v = algebra.create_vector(16, {HrrProperties.POSITIVE}, rng=rng)

    assert np.allclose(
        algebra.bind(algebra.binding_power(v, 2.2), algebra.binding_power(v, 3.3)),
        algebra.binding_power(v, 5.5),
    )


def test_fractional_binding_power_of_non_positive_vector_raises(rng):
    algebra = HrrAlgebra()
    v = algebra.bind(
        algebra.create_vector(16, {HrrProperties.POSITIVE}, rng=rng),
        HrrSign(-1, 1).to_vector(16),
    )
    assert algebra.sign(v).is_negative()
    with pytest.raises(ValueError):
        algebra.binding_power(v, 0.5)


class TestHrrSign:
    @pytest.mark.parametrize("sign", [HrrSign(1, 1), HrrSign(1, 0)])
    def test_positive(self, sign):
        assert sign.is_positive()
        assert not sign.is_negative()
        assert not sign.is_zero()
        assert not sign.is_indefinite()

    @pytest.mark.parametrize(
        "sign", [HrrSign(1, -1), HrrSign(-1, 1), HrrSign(-1, -1), HrrSign(-1, 0)]
    )
    def test_negative(self, sign):
        assert not sign.is_positive()
        assert sign.is_negative()
        assert not sign.is_zero()
        assert not sign.is_indefinite()

    def test_zero(self):
        sign = HrrSign(0, 0)
        assert not sign.is_positive()
        assert not sign.is_negative()
        assert sign.is_zero()
        assert not sign.is_indefinite()

    @pytest.mark.parametrize(
        "dc_sign, nyquist_sign", [(0, 1), (0, -1), (0.5, 0), (1, 2)]
    )
    def test_invalid(self, dc_sign, nyquist_sign):
        with pytest.raises(ValueError):
            HrrSign(dc_sign, nyquist_sign)

    @pytest.mark.parametrize(
        "sign,expected",
        [
            (HrrSign(0, 0), [0, 0, 0, 0]),
            (HrrSign(1, 1), [1, 0, 0, 0]),
            (HrrSign(1, 0), [1, 0, 0, 0]),
            (HrrSign(1, -1), [0, 1, 0, 0]),
            (HrrSign(-1, 1), [0, -1, 0, 0]),
            (HrrSign(-1, 0), [-1, 0, 0, 0]),
            (HrrSign(-1, -1), [-1, 0, 0, 0]),
            (HrrSign(0, 0), [0, 0, 0]),
            (HrrSign(1, 0), [1, 0, 0]),
            (HrrSign(-1, 0), [-1, 0, 0]),
        ],
    )
    def test_to_vector(self, sign, expected):
        assert HrrAlgebra().sign(expected) == sign  # sanity check
        assert np.allclose(sign.to_vector(len(expected)), expected)

    def test_equality(self):
        assert HrrSign(1, 1) == HrrSign(1, 1)
        assert HrrSign(-1, 1) == HrrSign(-1, 1)
        assert HrrSign(0, 0) == HrrSign(0, 0)
        assert HrrSign(1, -1) == HrrSign(1, -1)

        assert HrrSign(1, 1) != HrrSign(1, -1)
        assert HrrSign(1, 1) != HrrSign(0, 0)

        assert HrrSign(1, 1) != (1, 1)

    def test_repr(self):
        sign = HrrSign(-1, 1)
        assert eval(repr(sign)) == sign
